import logging
import os
from datetime import date, datetime, timedelta

import boto3
from airflow import DAG
from airflow.operators.bash import BashOperator
from airflow.operators.python import BranchPythonOperator, PythonOperator
from airflow.providers.amazon.aws.sensors.s3 import S3KeySensor
from airflow.providers.cncf.kubernetes.operators.kubernetes_pod import (
    KubernetesPodOperator,
)
from airflow.sensors.python import PythonSensor
from airflow.utils.trigger_rule import TriggerRule
from kubernetes.client import V1EnvVar

from mycroft.common.cross_acct_s3 import custom_input_s3_key_sensor
from mycroft.common.reason_queries.process_proc_hash_cert_daily_alert_dev_count_table import (
    process_daily_intermediate_table,
)
from mycroft.common.settings import get_es_urls
from mycroft.etl.move_data import move_last_export
from mycroft.generate_suggestions.generate_suggestions_using_delta_method import (
    create_rds_tables,
    empty_hash_sugg_gen,
    export_pool_data,
    gen_cert_sug,
    gen_hash_sug,
    gen_proc_rules_sug,
    gen_recommendations_from_s3,
    gen_wf_rec_stats,
    insert_recommendations_into_backend_postgres,
    insert_s3_cert_pool_intermediate_rds,
    insert_s3_hash_pool_intermediate_rds,
)

# Default airflow dag arguments
default_args = {
    "owner": "airflow",
    "email_on_failure": False,
    "email_on_retry": False,
    "retries": 3,
    "retry_delay": timedelta(minutes=5),
}

# Define the dag
dag = DAG(
    dag_id="generate_recommendations",
    description="Generates Reputation/Policy Recommendations",
    default_args=default_args,
    schedule_interval="5 0 * * *",  # schedule this task at 00:05
    start_date=datetime(2020, 11, 24),
    catchup=False,
    is_paused_upon_creation=False,
)

try:

    REGION = os.environ["REGION"]
    env_category = os.environ["STAGE"]
    # policy-sharing S3 bucket
    data_s3_bucket = os.environ["DATA_S3_BUCKET"]
    env = os.environ["ENV"]
    legacy_env = os.getenv("LEGACY_ENV", default="")
    intermediate_db_s3_bucket = os.environ["INTERMEDIATE_BUCKET_NAME"]
    # Use intermediate s3 bucket for storing query result only in UkPoP
    query_output_s3_bucket = os.getenv("QUERY_OUTPUT_BUCKET", intermediate_db_s3_bucket)
    generate_date = f"{date.today()}"
    generate_stage = "bootstrap"
    image_base = os.environ["ECR_IMAGE_BASE"]
    image_tag = os.environ["IMAGE_TAG"]
    namespace = os.environ["NAMESPACE"]
    hash_pool_complete_key = f"{generate_date}/rules_pool/proc_sha256hash/pool.csv"
    cert_pool_complete_key = f"{generate_date}/rules_pool/cert/pool.csv"
    # TODO: remove msc2 after switching pipeline to msc2.
    output_recommendations_key = f"{generate_date}/recommendations-msc2.csv"
    tool_pool_key = f"{generate_date}/rules_pool/it_tools/grouped_rules_pool.csv"
    metric_host = os.getenv("METRIC_HOST", "wavefront-proxy")
    metric_port = os.getenv("METRIC_PORT", "2878")
    # No Token is needed if we are using WF proxy URL, the proxy will take care
    wf_server = f"http://{metric_host}:{metric_port}"
    wf_source = f"airflow-{env_category}-{REGION}"

    flink_database = os.environ["FLINK_PRIMARY_INFRA_DB_NAME"]
    daily_intermediate_db_name = os.environ["DAILY_INTERMEDIATE_DB"]
    # Assume mosaic1 IAM role ARN
    msc1_role_arn = os.getenv("MOSAIC1_IAM_ARN", None)
except KeyError as e:
    logging.error(f"Failed to read environment variable {e}")


executor_config = {
    "KubernetesExecutor": {"image": f"{image_base}policysuggestion:{image_tag}"}
}

daily_update_intermediate_table = PythonOperator(
    task_id="update_intermediate_table",
    python_callable=process_daily_intermediate_table,
    dag=dag,
    provide_context=True,
    executor_config=executor_config,
    op_kwargs={
        "region": REGION,
        "legacy_env": legacy_env,
        "msc1_role_arn": msc1_role_arn,
        "query_output_s3_bucket": query_output_s3_bucket,
        "database_s3_bucket": intermediate_db_s3_bucket,
        "from_database": flink_database,
        "to_database": daily_intermediate_db_name,
    },
    retries=3,
)

sugg_gen_config = {
    "KubernetesExecutor": {
        "image": f"{image_base}policysuggestion:{image_tag}",
        "request_cpu": "2",
        "request_memory": "24Gi",
        "limit_memory": "24Gi",
    }
}

hash_sugg_gen = PythonOperator(
    task_id="hash_sugg_gen",
    python_callable=gen_hash_sug,
    retries=3,
    dag=dag,
    executor_config=sugg_gen_config,
    op_kwargs={
        "generate_date": generate_date,
        "stage": generate_stage,
    },
)

cert_sugg_gen = PythonOperator(
    task_id="cert_sugg_gen",
    python_callable=gen_cert_sug,
    retries=3,
    dag=dag,
    executor_config=sugg_gen_config,
    op_kwargs={
        "generate_date": generate_date,
        "stage": generate_stage,
    },
)

it_tool_rules_sug_gen = PythonOperator(
    task_id="it_tool_rules_sug_gen",
    python_callable=gen_proc_rules_sug,
    retries=3,
    dag=dag,
    executor_config=sugg_gen_config,
    op_kwargs={
        "generate_date": generate_date,
        "generation_type": "it_tools",
        "stage": generate_stage,
    },
)

for es_env, es_url in get_es_urls().items():
    es_dump_location = f"s3://{data_s3_bucket}/dump_es_{es_env}.json"
    env_vars = [
        V1EnvVar(name="ES_URL", value=es_url),
        V1EnvVar(name="S3_PATH", value=es_dump_location),
    ]
    dump_es = KubernetesPodOperator(
        namespace=namespace,
        image=f"{image_base}policy-suggest-dump-es:{image_tag}",
        image_pull_policy="Always",
        labels={"step": "1"},
        name=f"dump_es_{es_env}",
        task_id=f"dump_es_{es_env}",
        security_context={"runAsUser": 1000, "runAsGroup": 2000},
        env_vars=env_vars,
        get_logs=True,
        is_delete_operator_pod=True,
        dag=dag,
    )

    bash_command = f"""
    /opt/airflow/dags/extract_data_for_policy_org_reputation_override.sh {es_dump_location} {es_env} {data_s3_bucket} 2>&1 | tee /opt/airflow/logs/airflow_bash_operator_log.txt
    """

    extract_policy_org_override = BashOperator(
        task_id=f"extract_policy_org_override_{es_env}",
        bash_command=bash_command,
        dag=dag,
    )

    (
        dump_es
        >> extract_policy_org_override
        >> [
            hash_sugg_gen,
            cert_sugg_gen,
            it_tool_rules_sug_gen,
        ]
    )

extract_rds = PythonOperator(
    task_id="extract_rds",
    python_callable=move_last_export,
    dag=dag,
    executor_config=executor_config,
    op_kwargs={
        "region": REGION,
        "env": env,
        "legacy_env": legacy_env,
        "bucket": data_s3_bucket,
    },
)

# perm_rules_sug_gen = PythonOperator(
#     task_id="perm_rules_sug_gen",
#     python_callable=gen_proc_rules_sug,
#     retries=3,
#     dag=dag,
#     op_kwargs={
#         "generate_date": generate_date,
#         "generation_type": "perm_rule",
#         "stage": generate_stage,
#     }
# )

extract_rds >> [
    hash_sugg_gen,
    cert_sugg_gen,
    it_tool_rules_sug_gen,
]

daily_update_intermediate_table >> [
    hash_sugg_gen,
    cert_sugg_gen,
    it_tool_rules_sug_gen,
]

# Each region needs to generate the recommendations after the pool is built.
data_bucket = os.getenv("INTERMEDIATE_BUCKET_NAME")

# complete hash generation/merge with events after pool file is there.
hash_pool_complete_file_sensor = S3KeySensor(
    task_id="hash_pool_complete_file_sensor",
    bucket_key=hash_pool_complete_key,
    bucket_name=data_bucket,
    aws_conn_id=f"s3://{data_bucket}",
    timeout=10 * 60 * 60,
    poke_interval=120,
    dag=dag,
)

# complete cert generation/merge with events after pool file is there.
cert_pool_complete_file_sensor = S3KeySensor(
    task_id="cert_pool_complete_file_sensor",
    bucket_key=cert_pool_complete_key,
    bucket_name=data_bucket,
    aws_conn_id=f"s3://{data_bucket}",
    timeout=10 * 60 * 60,
    poke_interval=120,
    dag=dag,
)

gen_recommendations_from_s3_task = PythonOperator(
    task_id="gen_recommendations_from_s3_task",
    python_callable=gen_recommendations_from_s3,
    retries=3,
    dag=dag,
    executor_config=executor_config,
    op_kwargs={
        "s3_bucket": data_bucket,
        "hash_events_s3_key": f"{generate_date}/events_pool/proc_sha256hash/all_events.csv",
        "hash_pool_s3_key": hash_pool_complete_key,
        "existing_hash_rules_s3_key": f"{generate_date}/existing_rules/proc_sha256hash/hash_rule_list.csv",
        "cert_events_s3_key": f"{generate_date}/events_pool/cert/all_events.csv",
        "cert_pool_s3_key": cert_pool_complete_key,
        "existing_cert_rules_s3_key": f"{generate_date}/existing_rules/cert/cert_rule_list.csv",
        "it_tool_recommendation_key": f"{generate_date}/rules_pool/it_tools/recommendations.csv",
        "existing_it_tool_key": f"{generate_date}/existing_rules/it_tools/it_tools_rule_list.csv",
        "output_recommendations_key": output_recommendations_key,
    },
)


# Now only creates CSV file.
recommendations_into_backend_postgres = PythonOperator(
    task_id="recommendations_into_backend_postgres",
    python_callable=insert_recommendations_into_backend_postgres,
    retries=3,
    dag=dag,
    executor_config=executor_config,
    op_kwargs={
        "s3_bucket": data_bucket,
        "output_recommendations_key": output_recommendations_key,
        "hash_pool_key": hash_pool_complete_key,
        "cert_pool_key": cert_pool_complete_key,
        "it_tool_pool_key": tool_pool_key,
    },
)

gen_wf_rec_stats_task = PythonOperator(
    task_id="gen_wf_rec_stats_task",
    python_callable=gen_wf_rec_stats,
    retries=3,
    dag=dag,
    executor_config=executor_config,
    op_kwargs={
        "s3_bucket": data_bucket,
        "output_recommendations_key": output_recommendations_key,
        "wf_server": wf_server,
        "wf_source": wf_source,
    },
)

hash_sugg_gen >> hash_pool_complete_file_sensor
cert_sugg_gen >> cert_pool_complete_file_sensor
(
    [
        hash_pool_complete_file_sensor,
        cert_pool_complete_file_sensor,
        it_tool_rules_sug_gen,
    ]
    >> gen_recommendations_from_s3_task
    >> [
        gen_wf_rec_stats_task,
        recommendations_into_backend_postgres,
    ]
)

# in the US exclusively build the pool of all data, then export it to all regions.
if REGION == "us-east-1":
    if env_category == "int":
        # Only one region for kcdev01, kctest03
        regions = ["us-east-1", "eu-west-2"]
    elif env_category == "prd":
        # Create multiple region sensors
        regions = [
            "us-east-1",
            "eu-central-1",
            "ap-northeast-1",
            "ap-southeast-2",
            "eu-west-2",  # VDP environment
        ]
    else:
        regions = []

    """Create required RDS tables and clean up existing data.
    """
    rds_tables_create = PythonOperator(
        task_id="rds_tables_create",
        python_callable=create_rds_tables,
        retries=3,
        dag=dag,
        executor_config=executor_config,
    )

    """Run hash_query and cert_query in US-EAST-1 RDS
    Write result to intermediate S3 bucket of each region
    """
    intermediate_bucket_category = "prod" if env_category == "prd" else "dev"
    pop_env = "prd1ew2" if env_category == "prd" else "int1usw2"
    export_pool_data_task = PythonOperator(
        task_id="export_pool_data_task",
        python_callable=export_pool_data,
        retries=3,
        dag=dag,
        executor_config=executor_config,
        op_kwargs={
            "s3_bucket_list": [
                f"cdca-{intermediate_bucket_category}-{region_name}-intermediate-data"
                for region_name in regions
                if region_name != "eu-west-2"
            ]
            + [
                # UK Pop
                f"{pop_env}-policysuggestion-infra-intermediate-data-bucket"
            ],
            "hash_key": hash_pool_complete_key,
            "cert_key": cert_pool_complete_key,
        },
    )

    [hash_sugg_gen, cert_sugg_gen] >> rds_tables_create
    export_pool_data_task >> [
        hash_pool_complete_file_sensor,
        cert_pool_complete_file_sensor,
    ]

    """Reads hash_pool_s3_key and cert_pool_s3_key from intermediate S3 bucket of each region
    Insert both CSV files to RDS tables (hash_pool, cert_pool)
    """
    for sensor_region in regions:
        hash_pool_s3_key = f"{generate_date}/rules_pool/proc_sha256hash/hash_list_with_common_path_and_file_name.csv"
        cert_pool_s3_key = f"{generate_date}/rules_pool/cert/cert_list.csv"
        if sensor_region == "eu-west-2":
            # Create sensors for buckets in Mosaic2 infra (scargo).
            data_bucket = f"{pop_env}-policysuggestion-infra-intermediate-data-bucket"
            hash_pool_s3_file_sensor = S3KeySensor(
                task_id=f"hash_pool_s3_file_sensor_{sensor_region}",
                bucket_key=hash_pool_s3_key,
                bucket_name=data_bucket,
                aws_conn_id=f"s3://{data_bucket}",
                timeout=2 * 60 * 60,
                poke_interval=120,
                dag=dag,
            )
            cert_pool_s3_file_sensor = S3KeySensor(
                task_id=f"cert_pool_s3_file_sensor_{sensor_region}",
                bucket_key=cert_pool_s3_key,
                bucket_name=data_bucket,
                aws_conn_id=f"s3://{data_bucket}",
                timeout=2 * 60 * 60,
                poke_interval=120,
                dag=dag,
            )
        else:
            # Create sensors for buckets in Mosaic1 infra (cdc-analytic) requires cross account role.
            data_bucket = (
                f"cdca-{intermediate_bucket_category}-{sensor_region}-intermediate-data"
            )
            sts_client = boto3.client("sts")
            assumed_role_object = sts_client.assume_role(
                RoleArn=msc1_role_arn, RoleSessionName="AssumeMCS1Role"
            )
            credentials = assumed_role_object["Credentials"]
            hash_pool_s3_file_sensor = PythonSensor(
                task_id=f"hash_pool_s3_file_sensor_{sensor_region}",
                python_callable=custom_input_s3_key_sensor,
                op_kwargs={
                    "bucket_name": data_bucket,
                    "file_name": hash_pool_s3_key,
                    "access_key": credentials["AccessKeyId"],
                    "secret_key": credentials["SecretAccessKey"],
                    "session_token": credentials["SessionToken"],
                },
                timeout=2 * 60 * 60,
                poke_interval=120,
                dag=dag,
                executor_config=executor_config,
            )
            cert_pool_s3_file_sensor = PythonSensor(
                task_id=f"cert_pool_s3_file_sensor_{sensor_region}",
                python_callable=custom_input_s3_key_sensor,
                op_kwargs={
                    "bucket_name": data_bucket,
                    "file_name": cert_pool_s3_key,
                    "access_key": credentials["AccessKeyId"],
                    "secret_key": credentials["SecretAccessKey"],
                    "session_token": credentials["SessionToken"],
                },
                timeout=2 * 60 * 60,
                poke_interval=120,
                dag=dag,
                executor_config=executor_config,
            )

        # Read hash_pool_s3_key from regional S3 bucket and insert data into RDS of US-EAST-1
        hash_pool_s3_into_rds = PythonOperator(
            task_id=f"hash_pool_s3_into_rds_{sensor_region}",
            python_callable=insert_s3_hash_pool_intermediate_rds,
            retries=3,
            dag=dag,
            executor_config=executor_config,
            op_kwargs={
                "s3_bucket": data_bucket,
                "hash_pool_s3_key": hash_pool_s3_key,
                "cross_account_arn": msc1_role_arn
                if sensor_region != "eu-west-2"
                else None,
            },
        )

        (
            rds_tables_create
            >> hash_pool_s3_file_sensor
            >> hash_pool_s3_into_rds
            >> export_pool_data_task
        )

        # Read cert_pool_s3_key from regional S3 bucket and insert data into RDS of US-EAST-1
        cert_pool_s3_into_rds = PythonOperator(
            task_id=f"cert_pool_s3_into_rds_{sensor_region}",
            python_callable=insert_s3_cert_pool_intermediate_rds,
            retries=3,
            dag=dag,
            executor_config=executor_config,
            op_kwargs={
                "s3_bucket": data_bucket,
                "cert_pool_s3_key": cert_pool_s3_key,
                "cross_account_arn": msc1_role_arn
                if sensor_region != "eu-west-2"
                else None,
            },
        )

        (
            rds_tables_create
            >> cert_pool_s3_file_sensor
            >> cert_pool_s3_into_rds
            >> export_pool_data_task
        )
