import os
from datetime import datetime

from airflow import DAG
from airflow.operators.python import PythonOperator
from pyathena import connect

from mycroft.common.logging import get_logger

LOGLEVEL = os.environ.get("LOGLEVEL", "WARNING").upper()
logger = get_logger(LOGLEVEL)

try:
    STAGE = os.environ["STAGE"]
    FLINK_PRIMARY_INFRA_DB_NAME = f'{os.environ["FLINK_PRIMARY_INFRA_DB_NAME"]}_{STAGE}'
    DATABASE_LOCATION = os.environ["SOURCE_BUCKET"]
    BASELINE_POLICY_TABLE_NAME = os.environ["BASELINE_POLICY_TABLE_NAME"]
    BASELINE_POLICY_VIEW_NAME = os.environ["BASELINE_POLICY_VIEW_NAME"]
    BASELINE_REPUTATION_TABLE_NAME = os.environ["BASELINE_REPUTATION_TABLE_NAME"]
    BASELINE_DEVICE_MAPPING_TABLE_NAME = os.environ[
        "BASELINE_DEVICE_MAPPING_TABLE_NAME"
    ]
    REGION = os.environ["REGION"]
    logger.debug(f"FLINK_PRIMARY_INFRA_DB_NAME: {FLINK_PRIMARY_INFRA_DB_NAME}")
except KeyError as e:
    logger.error(f"Failed to read environment variable {e}")


def initialize_database_tables(
    baseline_db,
    baseline_db_location,
    baseline_policy_table,
    baseline_policy_view,
    baseline_reputation,
    baseline_device_mapping,
    region,
    **kwargs,
):
    """Create Glue database in policy-data-sharing S3 bucket for *_sugg_gen dags"""
    try:
        cursor = connect(
            s3_staging_dir=f"s3://{baseline_db_location}/",
            region_name=region,
        ).cursor()

        create_database_query = f"""
        CREATE DATABASE IF NOT EXISTS `{baseline_db}`
        LOCATION 's3://{baseline_db_location}'
        """
        logger.debug(f"Execute create baseline database query: {create_database_query}")
        cursor.execute(create_database_query)

        create_policy_table_query = f"""
        CREATE EXTERNAL TABLE {baseline_db}.{baseline_policy_table} (
          device_group_id int,
          org_id int,
          policy_control boolean,
          is_default boolean,
          priority int,
          name STRING,
          description STRING,
          admin_contact_id bigint,
          policy_info STRING,
          update_version int,
          priority_level int,
          change_timestamps STRING,
          analytics_profile STRING,
          analytics_version STRING,
          policy_info_v2 STRING
        ) PARTITIONED BY (env STRING) STORED AS PARQUET
          LOCATION 's3://{baseline_db_location}/policy/device_group/' tblproperties ("parquet.compression" = "gzip");
        """
        logger.debug(f"Execute create policy table query: {create_policy_table_query}")
        cursor.execute(create_policy_table_query)
        scan_policy_table = f"""
        MSCK REPAIR TABLE {baseline_db}.{baseline_policy_table};
        """
        cursor.execute(scan_policy_table)

        create_policy_device_mapping_table_query = f"""
        CREATE EXTERNAL TABLE {baseline_db}.{baseline_device_mapping} (
          device_group_id int,
          device_id bigint,
          msm_group_id int
        ) PARTITIONED BY (env STRING) STORED AS PARQUET
          LOCATION 's3://{baseline_db_location}/policy/device_group_device/' tblproperties ("parquet.compression" = "gzip");
        """
        logger.debug(
            f"Execute create policy device mapping table query: {create_policy_device_mapping_table_query}"
        )
        cursor.execute(create_policy_device_mapping_table_query)
        scan_device_mapping_table = f"""
        MSCK REPAIR TABLE {baseline_db}.{baseline_device_mapping};
        """
        cursor.execute(scan_device_mapping_table)

        create_policy_org_reputation_override_table_query = f"""
        CREATE EXTERNAL TABLE `{baseline_db}.{baseline_reputation}` (
          authlisttype string,
          createtime string,
          createdby string,
          csraction string,
          description string,
          entryid string,
          fileid string,
          id string,
          lastmodification string,
          lastupdatedtime string,
          orgid string,
          sourcetype string,
          type string,
          updateversion string,
          value1 string,
          value2 string)
        PARTITIONED BY (env string)
        ROW FORMAT SERDE 'org.apache.hadoop.hive.serde2.OpenCSVSerde'
        WITH SERDEPROPERTIES (
            'escapeChar'='\\\\',
            'quoteChar'='\\"',
            'separatorChar'='\\t')
        STORED AS INPUTFORMAT 'org.apache.hadoop.mapred.TextInputFormat'
        OUTPUTFORMAT 'org.apache.hadoop.hive.ql.io.HiveIgnoreKeyTextOutputFormat'
        LOCATION 's3://{baseline_db_location}/orgrepoverride_table';
        """
        logger.debug(
            f"Execute create policy reputation query: {create_policy_org_reputation_override_table_query}"
        )
        cursor.execute(create_policy_org_reputation_override_table_query)
        scan_reputation_table = f"""
        MSCK REPAIR TABLE {baseline_db}.{baseline_reputation};
        """
        cursor.execute(scan_reputation_table)

        create_policy_view_query = f"""
        CREATE VIEW {baseline_db}.{baseline_policy_view} AS
        with t AS (with t1 AS
            (SELECT *,
                 CAST(json_extract(policy_info_v2,
                 '$.rules') AS ARRAY(MAP(VARCHAR, JSON))) AS rules_arr
            FROM {baseline_db}.{baseline_policy_table})
            SELECT device_group_id, org_id, name, env,
                CAST(json_extract(rules['application'], '$.value') AS VARCHAR) AS application_value,
                CAST(json_extract(rules['application'], '$.type') AS VARCHAR) AS application_type,
                CAST(rules['operation'] AS VARCHAR) AS operation,
                CAST(rules['action'] AS VARCHAR) AS action
            FROM t1
            CROSS JOIN UNNEST(rules_arr) AS t(rules))
        SELECT *
        FROM t
        WHERE application_type = 'NAME_PATH'
        """
        logger.debug(f"Execute create policy view query: {create_policy_view_query}")
        cursor.execute(create_policy_view_query)
    except Exception as e:
        logger.error(
            f"Failed to initialize, baseline DB={baseline_db}, DB location={baseline_db_location}, "
            f"policy table={baseline_policy_table}, policy view={baseline_policy_view}, "
            f"reputation table={baseline_reputation}, device mapping table={baseline_device_mapping}, region={region}"
            f"Exception: {str(e)}"
        )


with DAG(
    dag_id="initialization_baseline_db",
    description="Initialize database and tables for all suggestion generation.",
    default_args={
        "owner": "airflow",
        "project": "policy_automation",
    },
    schedule_interval=None,
    start_date=datetime(2021, 2, 26),
    is_paused_upon_creation=False,
) as dag:
    initialization = PythonOperator(
        task_id="initialize_baseline_db",
        provide_context=True,
        python_callable=initialize_database_tables,
        op_kwargs={
            "baseline_db": FLINK_PRIMARY_INFRA_DB_NAME,
            "baseline_db_location": DATABASE_LOCATION,
            "baseline_policy_table": BASELINE_POLICY_TABLE_NAME,
            "baseline_policy_view": BASELINE_POLICY_VIEW_NAME,
            "baseline_reputation": BASELINE_REPUTATION_TABLE_NAME,
            "baseline_device_mapping": BASELINE_DEVICE_MAPPING_TABLE_NAME,
            "region": REGION,
        },
    )
