import logging
import os
from datetime import datetime, timedelta

import boto3
from airflow import DAG
from airflow.operators.python import PythonOperator

# Default airflow dag arguments
default_args = {
    "owner": "airflow",
    "project": "policy_automation",
    "email_on_failure": False,
    "email_on_retry": False,
    "retries": 3,
    "retry_delay": timedelta(minutes=1),
}

region = os.getenv("REGION", "us-west-2")
stage = os.getenv("STAGE")
start_time = datetime.now()
database_identifier = os.getenv("DATABASE_IDENTIFIER")
bucket_name = os.getenv("SOURCE_BUCKET")
iam_role_arn = os.getenv("IAM_ROLE_ARN")
kms_key_id = os.getenv("KMS_KEY_ID")
rds_client = boto3.client("rds", region_name=region)
s3_client = boto3.client("s3", region_name=region)


def create_rds_snapshot_data(start_time, **kwargs):
    logging.info("start_time = %s", start_time)
    snapshots = rds_client.describe_db_cluster_snapshots(
        DBClusterIdentifier=database_identifier, IncludeShared=True, IncludePublic=True
    )["DBClusterSnapshots"]
    if snapshots:

        def create_time(snapshots):
            return snapshots["SnapshotCreateTime"]

        snapshotArn = sorted(snapshots, key=create_time, reverse=True)[0][
            "DBClusterSnapshotArn"
        ]
        logging.info("Snapshot ARN: %s" + snapshotArn)
        export_task_identifier = database_identifier + start_time.strftime(
            "%Y-%m-%d-%H-%M-%S"
        )
        logging.info("Export task identifier = %s", export_task_identifier)
        export_job = rds_client.start_export_task(
            ExportTaskIdentifier=export_task_identifier,
            SourceArn=snapshotArn,
            S3BucketName=bucket_name,
            IamRoleArn=iam_role_arn,
            KmsKeyId=kms_key_id,
        )
        print("Export Task Identifier: " + export_job["ExportTaskIdentifier"])
        if "FailureCause" in export_job:
            print("Export job failed. Reason: " + export_job["FailureCause"])
            exit(1)


# Define DAG
dag = DAG(
    dag_id="rds_snap_export",
    description="Creates RDS Snapshots",
    default_args=default_args,
    schedule_interval="00 15 * * 0,3",  # schedule this task at 22:05
    start_date=datetime(2022, 2, 17),
    catchup=False,
    is_paused_upon_creation=False,
)
rds_snap_export = PythonOperator(
    task_id="rds_snap_export",
    python_callable=create_rds_snapshot_data,
    dag=dag,
    op_kwargs={
        "start_time": start_time,
    },
    retries=3,
)
